/***************************************************************************
                          AVindex.c  -  description
                             -------------------
   Functions to store and access the index of current AV stream.
                             -------------------
    begin                : Thu Jan 26 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define DISPLAY_AVINDEX*/
#ifdef DISPLAY_AVINDEX
# include <stdio.h>
#endif

#include "AVindex.h"

/* index memory = 24bytes * AVINDEX_SIZE = 12KB */
/* AVINDEX_SIZE must be power of 2 */
#define AVINDEX_SIZE    512
#define AVINDEX_NEXT(i) (i+1)&(AVINDEX_SIZE-1)

typedef struct {
   FILE_POS f_pos;
   CLK_VAL  v_pos, a_pos;
   u32      v_bck, a_bck;
} INDEX_ENTRY;

static INDEX_ENTRY *index = NULL;
static int          start =  0,
                    end   = -1;

/* Returns the number of bytes of memory needed by the AVindex. */
size_t AVindex_mem_needed()
{
   return (sizeof(INDEX_ENTRY)*AVINDEX_SIZE);
}

/* Uses the memory pointed to by mem, for the AVindex. */
void AVindex_set_mem(void *mem)
{
#ifdef DISPLAY_AVINDEX
   if ( (!mem)&&(index)&&(end >= 0) ) {
      int i = 0;
      printf("     |Video           |  |Audio           |  |File      |\n"
             "     |================|  |================|  |==========|\n");
      do {
         if (i)
            start = AVINDEX_NEXT(start);
         printf("%3d> |%7.2lfs,%7u|  |%7.2lfs,%7u|  |0x%08X|\n", i++,
                (double)index[start].v_pos / (double)MPEG2_SYS_CLK_FREQ,
                index[start].v_bck,
                (double)index[start].a_pos / (double)MPEG2_SYS_CLK_FREQ,
                index[start].a_bck,
                (u32)index[start].f_pos);
      } while (start != end);
   }
#endif
   index = (INDEX_ENTRY *)mem;
   start =  0;
   end   = -1;
}

/* Clear the index, and put the given data at the begining of
    the index.
      INPUT: f_pos - File position of the first entry.
             v_pos - Time position of the video stream.
             a_pos - Time position of the audio stream.
             v_bck - Bytes behind f_pos, where v_pos occurred.
             a_bck - Bytes behind f_pos, where a_pos occurred. */
void AVindex_init(FILE_POS f_pos,
                  CLK_VAL  v_pos, CLK_VAL a_pos,
                  u32      v_bck, u32     a_bck)
{
   if (v_bck > 1)
      f_pos -= (FILE_POS)v_bck;
   else if (a_bck > 1)
      f_pos -= (FILE_POS)a_bck;

   start = end = 0;
   index[0].f_pos = f_pos;
   index[0].v_pos = v_pos; index[0].a_pos = a_pos;
   index[0].v_bck = v_bck; index[0].a_bck = a_bck;
#ifdef DISPLAY_AVINDEX
   printf("     |Video           |  |Audio           |  |File      |\n"
          "     |================|  |================|  |==========|\n");
   printf("INI> |%7.2lfs,%7u|  |%7.2lfs,%7u|  |0x%08X| %3d %3d\n",
          (double)v_pos / (double)MPEG2_SYS_CLK_FREQ, v_bck,
          (double)a_pos / (double)MPEG2_SYS_CLK_FREQ, a_bck,
          (u32)f_pos, start, end);
#endif
}

/* Add a new entry to the index.
      INPUT: f_pos - File position of the new entry.
             v_pos - Time position of the video stream.
             a_pos - Time position of the audio stream.
             v_bck - Bytes behind f_pos, where v_pos occurred.
             a_bck - Bytes behind f_pos, where a_pos occurred. */
void AVindex_add(FILE_POS f_pos,
                 CLK_VAL  v_pos, CLK_VAL a_pos,
                 u32      v_bck, u32     a_bck)
{
   end = AVINDEX_NEXT(end);
   if (start == end)
      start = AVINDEX_NEXT(start);

   if (v_bck > 1)
      f_pos -= (FILE_POS)v_bck;
   else if (a_bck > 1)
      f_pos -= (FILE_POS)a_bck;

   index[end].f_pos = f_pos;
   index[end].v_pos = v_pos; index[end].a_pos = a_pos;
   index[end].v_bck = v_bck; index[end].a_bck = a_bck;
#ifdef DISPLAY_AVINDEX
   printf("ADD> |%7.2lfs,%7u|  |%7.2lfs,%7u|  |0x%08X| %3d %3d\n",
          (double)v_pos / (double)MPEG2_SYS_CLK_FREQ, v_bck,
          (double)a_pos / (double)MPEG2_SYS_CLK_FREQ, a_bck,
          (u32)f_pos, start, end);
#endif
}

/* Get an entry from the index.
      INPUT:  v_pos - Video position to search for. NULL if not needed.
              a_pos - Audio position to search for. NULL if not needed.
      OUTPUT: f_pos - File position of the requested entry.
              v_pos - Time position of the video stream.
              a_pos - Time position of the audio stream.
              v_bck - Bytes behind f_pos, where v_pos occurred.
              a_bck - Bytes behind f_pos, where a_pos occurred.
      RETURN: Success   - non-zero value
              Not found - 0 */
int AVindex_get(FILE_POS *f_pos,
                CLK_VAL  *v_pos, CLK_VAL *a_pos,
                u32      *v_bck, u32     *a_bck)
{
   int next, vidx, aidx;

#ifdef DISPLAY_AVINDEX
   printf("GET> |%7.2lfs        |  |%7.2lfs        | %3d %3d\n",
# ifndef _NO_VIDEO_
          (double)(*v_pos) / (double)MPEG2_SYS_CLK_FREQ,
# else
          0.0,
# endif
# ifndef _NO_AUDIO_
          (double)(*a_pos) / (double)MPEG2_SYS_CLK_FREQ,
# else
          0.0,
# endif
          start, end);
#endif
   if (end < 0)
      return 0;   /* index is empty */

   if ( (v_pos)&&(*v_pos >= index[start].v_pos) ) {
      for (vidx  = start,              next = AVINDEX_NEXT(start);
           vidx != end;
           vidx  = AVINDEX_NEXT(vidx), next = AVINDEX_NEXT(next)) {
         if ( (*v_pos >= index[vidx].v_pos)&&
              (*v_pos <  index[next].v_pos) )
            break;
      }
   }
   else
      vidx = -1;

   if ( (a_pos)&&(*a_pos >= index[start].a_pos) ) {
      for (aidx  = start,              next = AVINDEX_NEXT(start);
           aidx != end;
           aidx  = AVINDEX_NEXT(aidx), next = AVINDEX_NEXT(next)) {
         if ( (*a_pos >= index[aidx].a_pos)&&
              (*a_pos <  index[next].a_pos) )
            break;
      }
   }
   else
      aidx = -1;

   if ( (vidx < 0)&&(aidx < 0) )
      return 0;
   else if ( (vidx < 0)||
             ( (aidx >= 0)&&(vidx != aidx)&&
               (index[aidx].f_pos < index[vidx].f_pos) ) )
      vidx = aidx;
#ifdef DISPLAY_AVINDEX
   printf("GOT> |%7.2lfs,%7u|  |%7.2lfs,%7u|  |0x%08X|  %3d\n",
          (double)index[vidx].v_pos / (double)MPEG2_SYS_CLK_FREQ,
          index[vidx].v_bck,
          (double)index[vidx].a_pos / (double)MPEG2_SYS_CLK_FREQ,
          index[vidx].a_bck,
          (u32)index[vidx].f_pos, vidx);
#endif
   if (f_pos)
      *f_pos = index[vidx].f_pos;
   if (v_pos)
      *v_pos = index[vidx].v_pos;
   if (v_bck)
      *v_bck = index[vidx].v_bck;
   if (a_pos)
      *a_pos = index[vidx].a_pos;
   if (a_bck)
      *a_bck = index[vidx].a_bck;
   return 1;
}

#if 0
def __DVD_INPUT__
/* Get an entry from the index.
      INPUT:  f_pos - File position to search for.
      OUTPUT: f_pos - File position of the requested entry.
              v_pos - Time position of the video stream.
              a_pos - Time position of the audio stream.
              v_bck - Bytes behind f_pos, where v_pos occurred.
              a_bck - Bytes behind f_pos, where a_pos occurred.
      RETURN: Success   - non-zero value
              Not found - 0 */
int AVindex_find_pos(FILE_POS *f_pos,
                     CLK_VAL  *v_pos, CLK_VAL *a_pos,
                     u32      *v_bck, u32     *a_bck)
{
   int idx;

   if ( (end < 0)||(!f_pos) )
      return 0;   /* index is empty */

#ifdef DISPLAY_AVINDEX
   printf("FIND>|                |  |                |  |0x%08X|\n",
          (u32)(*f_pos));
#endif

   for (idx = start; idx != end; idx = AVINDEX_NEXT(idx)) {
      if (*f_pos <= index[idx].f_pos)
         break;
   }

#ifdef DISPLAY_AVINDEX
   printf("GOT> |%7.2lfs,%7u|  |%7.2lfs,%7u|  |0x%08X|\n",
          (double)index[idx].v_pos / (double)MPEG2_SYS_CLK_FREQ,
          index[idx].v_bck,
          (double)index[idx].a_pos / (double)MPEG2_SYS_CLK_FREQ,
          index[idx].a_bck,
          (u32)index[idx].f_pos);
#endif
   if (f_pos)
      *f_pos = index[idx].f_pos;
   if (v_pos)
      *v_pos = index[idx].v_pos;
   if (v_bck)
      *v_bck = index[idx].v_bck;
   if (a_pos)
      *a_pos = index[idx].a_pos;
   if (a_bck)
      *a_bck = index[idx].a_bck;
   return 1;
}
#endif

#undef AVIINDEX_SIZE
#undef AVINDEX_NEXT

