/***************************************************************************
                          MPEG2_SYS.h  -  description
                             -------------------
   MPEG2 module function prototypes, and structures.
                             -------------------
    begin                : Sun Aug 8 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __MPEG2_SYS_H__
#define __MPEG2_SYS_H__

#include "common/typedefs.h"
#include "common/Mutex.h"
#include "common/CondVar.h"
#include "common/Thread.h"
#include "common/File.h"
#include "TS_PKT.h"
#include "PMT_SECTION.h"


#ifndef _NO_VIDEO_
# define VIDEO_DECODERS           3
#else
# define VIDEO_DECODERS           0
#endif
#ifndef _NO_AUDIO_
# define AUDIO_DECODERS           2
#else
# define AUDIO_DECODERS           0
#endif
#define  MPEG2_SYS_PIDS       (VIDEO_DECODERS+AUDIO_DECODERS+2)
#define  BUFFER_SIZE          (  32*1024)
#define  MPEG2_SYS_BUF_PKTS   (BUFFER_SIZE/TS_PKT_SIZE)
#define  NUM_BUFFERS          ((512*1024)/BUFFER_SIZE)


/*#define MPEG2_SYS_CLK64*/
#ifdef MPEG2_SYS_CLK64
# define MPEG2_SYS_CLK_FREQ  27000000
#else
# define MPEG2_SYS_CLK_FREQ     22500
#endif
#define  MPEG2_SYS_SLOW_DOWN        1

/* Use this macro to initialize a MPEG2_PSI_PES_MODULE structure. */
#define UNUSED_PSI_PES_MODULE \
{\
   (MPEG2_PSI_PES_MODULE_stream_init)NULL, \
   (MPEG2_PSI_PES_MODULE_process_pkt)NULL, \
   {\
      (MPEG2_MODULE_mem_needed)NULL, \
      (MPEG2_MODULE_init)NULL, \
      (MPEG2_MODULE_done)NULL, \
      (MPEG2_MODULE_set_offset)NULL, \
      (MPEG2_MODULE_discontinuity)NULL, \
      NULL \
   }\
}

/* Use this macro to initialize a MPEG2_SYS_MODULE structure. */
#define UNUSED_SYS_MODULE \
{\
   (MPEG2_SYS_MODULE_psi_pes_module)NULL, \
   (MPEG2_SYS_MODULE_process_pkt)NULL, \
   {\
      (MPEG2_MODULE_mem_needed)NULL, \
      (MPEG2_MODULE_init)NULL, \
      (MPEG2_MODULE_done)NULL, \
      (MPEG2_MODULE_set_offset)NULL, \
      (MPEG2_MODULE_discontinuity)NULL, \
      NULL \
   }\
}


#ifdef __cplusplus
extern "C" {
#endif

#ifdef MPEG2_SYS_CLK64
typedef s64 CLK_VAL;
#else
typedef s32 CLK_VAL;
#endif

/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
typedef struct {
   u32     command;
   CLK_VAL a_sync, 
           a_pos, v_pos,
           a_skip, v_skip;
} SYS_DATA;

/* Returns the number of bytes of memory needed by the module. */
typedef size_t (*MPEG2_MODULE_mem_needed)();

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
typedef void   (*MPEG2_MODULE_init)(void *, SYS_DATA *);

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
typedef void   (*MPEG2_MODULE_done)(void *);

/* Gives the module the offset value needed to obtain the program clock value.
      INPUT: void *  - Pointer to the memory provided for the module.
             CLK_VAL - offset value */
typedef void   (*MPEG2_MODULE_set_offset)(void *, CLK_VAL);

/* Alerts the module to a discontinuity in the data stream.
      INPUT: void * - Pointer to the memory provided for the module. */
typedef void   (*MPEG2_MODULE_discontinuity)(void *);

typedef struct {
   /* Pointers to common module functions. */
   MPEG2_MODULE_mem_needed    mem_needed;
   MPEG2_MODULE_init          init;
   MPEG2_MODULE_done          done;
   MPEG2_MODULE_set_offset    set_offset;
   MPEG2_MODULE_discontinuity discontinuity;

   void *mem;  /* Pointer to memory provided for the module. */
} MPEG2_MODULE;

  
/**************************************************************************
   COMMON PSI and PES MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the PSI_SECTIONs
       or PES_PKTs they produce.
 **************************************************************************/
/* Initialize everything needed to handle the given elementary stream.
     The PES module will handle the elementary stream (if able) until the
     MPEG2_MODULE_done function is called, or this function is called
     again.  PSI modules should always return 0.
      INPUT: void *        - Pointer to the memory provided for the module.
             PMT_SECTION * - Pointer to the PMT data.
             int           - Index, in elem array, of the elementary stream
                               that needs to be handled.
      RETURN: Module is ready to handle
                the stream.               - Non-zero value.
              Module cannot handle the
                stream.                   - 0 */
typedef int (*MPEG2_PSI_PES_MODULE_stream_init)(void *, PMT_SECTION *, int);

/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PSI_SECTION or PES_PKT that needs
                        handling. */
typedef void (*MPEG2_PSI_PES_MODULE_process_pkt)(void *, void *);

typedef struct {
   /* Pointers to common PSI or PES module functions. */
   MPEG2_PSI_PES_MODULE_stream_init stream_init;
   MPEG2_PSI_PES_MODULE_process_pkt process_pkt;

   MPEG2_MODULE common;
} MPEG2_PSI_PES_MODULE;


/*************************************************************************
   PAT MODULE FUNCTIONS
     These functions are used by the MPEG2 TS decoder to get the PAT data.
       The PAT data is extracted from the PSI_SECTIONs sent by a system
       module.
 *************************************************************************/
/* Adds an entry to the PAT, using the given values.  This function is
     used when there is no PAT data in the stream.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Program number of the new entry. Any value greater
                        than 0 is valid.
             u16    - Program Map PID of the new entry. Any unused user
                        defined PID value is valid. */
typedef void (*MPEG2_PAT_MODULE_add_entry)(void *, u16, u16);

/* Returns the PID of the TS_PKTs that contains the PMT data of the
     specified program number.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Program number to search for.
      RETURN: Program found     - Valid user defined PID value.
              Program not found - PID value that is >= TS_PKT_NUL_ID. */
typedef u16  (*MPEG2_PAT_MODULE_get_pid)(void *, u16);

/* Returns the program number at the specified location, in the
     access table.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Table index to search for.
      RETURN: Program found     - Valid program number.
              Program not found - 0 */
typedef u16  (*MPEG2_PAT_MODULE_get_program)(void *, u16);


/*************************************************************************
   PMT MODULE FUNCTIONS
     These functions are used by the MPEG2 TS decoder to get the PMT data.
       The PAT data is extracted from the PSI_SECTIONs sent by a system
       module.
 *************************************************************************/
/* Returns a non-zero value if the PMT data has changed, since the last
     time this function was called.
      INPUT: void * - Pointer to the memory provided for the module. */
typedef int          (*MPEG_PMT_MODULE_PMT_changed)(void *);

/* Adds an entry to the PMT, using the given values.  This function is
     used when there is no PMT data in the stream.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Program number of the new entry. Same value used
                        in the MPEG2_PAT_MODULE_add_entry() function.
             u16    - PID of the PCR.
             u16    - PID of the video stream.
             u8     - One of the STREAM_TYPE_* values defined
                        in PMT_SECTION.h
             u16    - PID of the audio stream. 
             u8     - One of the STREAM_TYPE_* values defined
                        in PMT_SECTION.h */
typedef void         (*MPEG2_PMT_MODULE_add_entry)(void *, u16, u16, u16, u8, u16, u8);

/* Returns a pointer to the PMT_SECTION for the current program.
      INPUT: void * - Pointer to the memory provided for the module.
      RETURN: The whole PMT_SECTION is stored - Pointer to a PMT_SECTION.
              No PMT_SECTION is stored        - NULL */
typedef PMT_SECTION *(*MPEG_PMT_MODULE_get_PMT)(void *);
                 

/**************************************************************************
   SYSTEM MODULE FUNCTIONS
     These functions are used by the MPEG2 TS decoder to extract the PSI or
       PES data from the corresponding TS_PKTs, and create PSI_SECTIONs
       or PES_PKTs for the PSI or PES modules.
 **************************************************************************/
/* Provides the PSI or PES module that will handle the PSI_SECTIONs or
     PES_PKTs.
      INPUT: void *                 - Pointer to the memory provided for
                                        the module.
             MPEG2_PSI_PES_MODULE * - Pointer to a PSI or PES module. */
typedef void (*MPEG2_SYS_MODULE_psi_pes_module)(void *, MPEG2_PSI_PES_MODULE *);

/* Processes the packet handled by the module.
      INPUT: void *   - Pointer to the memory provided for the module.
             TS_PKT * - Pointer to packet that needs handling. */
typedef void (*MPEG2_SYS_MODULE_process_pkt)(void *, TS_PKT *);

typedef struct {
   /* Pointers to SYSTEM module functions. */
   MPEG2_SYS_MODULE_psi_pes_module psi_pes_module;
   MPEG2_SYS_MODULE_process_pkt    process_pkt;
   
   MPEG2_MODULE common;
} MPEG2_SYS_MODULE;


typedef struct {
   CLK_VAL jump_offset,
           offset;            /* Offset value needed to obtain the program
                                   clock value. */
#ifdef MPEG2_SYS_CLK64
   CLK_VAL last_PCR;          /* Last PCR value encountered. */
#endif
       
   /* Pointer to the PAT module functions. */
   MPEG2_PAT_MODULE_add_entry   add_PAT_entry;
   MPEG2_PAT_MODULE_get_pid     get_pid;
   MPEG2_PAT_MODULE_get_program get_program;

   /* Pointers to the PMT module functions */   
   MPEG_PMT_MODULE_PMT_changed PMT_changed;
   MPEG2_PMT_MODULE_add_entry  add_PMT_entry;
   MPEG_PMT_MODULE_get_PMT     get_PMT;
   
   /* The PSI or PES modules that are used by the corresponding
        SYSTEM modules. */
   MPEG2_PSI_PES_MODULE     psi_pes[MPEG2_SYS_PIDS];

   /* The SYSTEM modules that handle the corresponding PIDs. */
   MPEG2_SYS_MODULE         module[MPEG2_SYS_PIDS];

   u16 PID[MPEG2_SYS_PIDS],   /* Active PID list. Use values >= TS_PKT_NUL_ID
                                   if you don't need to use all
                                   MPEG2_SYS_PIDS entries. Make sure one of
                                   the active PIDs is 0 (PAT). */
       PAT_idx,               /* Index value, in psi_pes array, of the
                                   PAT module. */
       PMT_idx,               /* Index value, in psi_pes array, of the
                                   PMT module. */
       PES_idx,               /* Index value, in module array, of any
                                   PES module. */
       program_number;        /* Program to be played. */
   
   MUTEX    m;
   CONDVAR  cv;
   SYS_DATA data;

#ifdef _RECORD_STATS_
   u32    stream_bytes, /* Number of bytes in the stream */
          pkt_cnt;      /* Number of packets in the stream */
   double decode_time;  /* Time (in secs) needed to decode stream */
#endif

   u8  *mem, *buf, *last_buf;
} MPEG2_SYS;
extern MPEG2_SYS sys;


typedef struct {
   u8     *buf;
   u32     bytes_read,
           si, fi;
   CLK_VAL pos;
#ifdef _RECORD_STATS_
   u32     pkt_cnt;  /* Number of packets in the stream */
#endif
} MPEG2_SYS_ARG;


/* Returns the offset needed to sync the system clock with the program clock.
      INPUT: PCR - Program clock reference value.
      RETURN: offset value. */
CLK_VAL MPEG2_get_offset(CLK_VAL PCR);

/* Returns the value of the program clock.
      INPUT: offset - Value needed to sync the system clock with the
                        program clock.
      RETURN: Current program clock value. */
CLK_VAL MPEG2_prog_clk(CLK_VAL offset);

/* Allocates, and initializes memory for all modules.
      RETURN: Success   - Non zero value.
              No memory - Zero. */
int  MPEG2_SYS_alloc();

/* Cleans up, and frees the memory allocated in MPEG2_SYS_alloc(). */
void MPEG2_SYS_free();

/* Uses the player function to play the content in the given file.
      INPUT: fp     - Points to the file that will be played.
             player - Function that can decode the content in
                        given file.
      RETURN: Success - A non-zero value
              Failure - 0 */
int  MPEG2_SYS_Play(FILE_PTR fp, THREAD_FUNC player);

void     MPEG2_SYS_modules_init();
void     MPEG2_SYS_modules_done();
void     MPEG2_SYS_decoder_pause();
FILE_POS MPEG2_SYS_decoder_jump(FILE_PTR       fp, 
                                MPEG2_SYS_ARG *v_arg, 
                                MPEG2_SYS_ARG *a_arg);
void     MPEG2_SYS_discontinuity_alert(MPEG2_SYS_ARG *v_arg,
                                       MPEG2_SYS_ARG *a_arg);
void     MPEG2_SYS_reset_PIDs();

#ifdef __cplusplus
};
#endif

#endif
