/***************************************************************************
                          PES_module.c  -  description
                             -------------------
   PES SYSTEM module functions.
                             -------------------
    begin                : Sat Sep 4 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <string.h>
#include <stdlib.h>

#include "PES_PKT.h"
#include "PES_module.h"


/* PRIVATE STRUCTURES */
typedef struct {
   MPEG2_PSI_PES_MODULE *pes_module;
   PES_PKT               pes;
   u32                   wait;
} PES_module_data;


/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t PES_SYS_MODULE_mem_needed()
{
   return sizeof(PES_module_data);
}

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void PES_SYS_MODULE_init(void *mem, SYS_DATA *sys_data)
{
   PES_module_data *data = (PES_module_data *)mem;

   data->wait      = 1; /* wait for start indicator */
   data->pes.state = 0;
   if (data->pes_module->common.init)
      data->pes_module->common.init(data->pes_module->common.mem, sys_data);
}

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void PES_SYS_MODULE_done(void *mem)
{
   PES_module_data *data = (PES_module_data *)mem;

   data->wait      = 1; /* wait for start indicator */
   data->pes.state = 0;
   if (data->pes_module->common.done)
      data->pes_module->common.done(data->pes_module->common.mem);
}

/* Gives the module the offset value needed to obtain the program clock value.
      INPUT: void *  - Pointer to the memory provided for the module.
             CLK_VAL - offset value */
void   PES_SYS_MODULE_set_offset(void *mem, CLK_VAL offset)
{
   PES_module_data *data = (PES_module_data *)mem;

   if (data->pes_module->common.set_offset)
      data->pes_module->common.set_offset(data->pes_module->common.mem, offset);
}

/* Alerts the module to a discontinuity in the data stream.
      INPUT: void * - Pointer to the memory provided for the module. */
void PES_SYS_MODULE_discontinuity(void *mem)
{
   PES_module_data *data = (PES_module_data *)mem;

   data->wait      = 1; /* wait for start indicator */
   data->pes.state = 0;
   if (data->pes_module->common.discontinuity)
      data->pes_module->common.discontinuity(data->pes_module->common.mem);
}


/**************************************************************************
   SYSTEM MODULE FUNCTIONS
     These functions are used by the MPEG2 TS decoder to extract the PSI
       data from the corresponding TS_PKTs, and create PSI_SECTIONs for
       the PSI mdules.
 **************************************************************************/
/* Provides the PES module that will handle the PES_PKTs.
      INPUT: void *                 - Pointer to the memory provided for
                                        the module.
             MPEG2_PSI_PES_MODULE * - Pointer to a PES module. */
void PES_SYS_MODULE_psi_pes_module(void *mem, MPEG2_PSI_PES_MODULE *pes)
{
   PES_module_data *data = (PES_module_data *)mem;

   data->pes_module = pes;
}

/* Processes the packet handled by the module.
      INPUT: void *   - Pointer to the memory provided for the module.
             TS_PKT * - Pointer to packet that needs handling. */
void PES_SYS_MODULE_process_pkt(void *mem, TS_PKT *ts)
{
   PES_module_data *data = (PES_module_data *)mem;

   if ( (ts->payload_unit_start_indicator)||(data->pes.state) ) {
      if (PES_PKT_read(ts->payload, ts->payload_len32, &(data->pes)))
         data->wait = 0;
      else {
         data->wait = 1;
         return;
      }
   }
   else if (data->wait)
      return;  /* waiting for start indicator */
   else {
      data->pes.payload_len16 = (u16)ts->payload_len32;
      data->pes.payload       = ts->payload;
   }

   /* send PES packet payload on to the PES module */
   data->pes_module->process_pkt(data->pes_module->common.mem, &(data->pes));
}

