/***************************************************************************
                          SEQ_HEADER.c  -  description
                             -------------------
   Stuff needed for handling MPEG2 sequence headers.
                             -------------------
    begin                : Fri Sep 17 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#if defined(_M2V_TRACE_) || defined(_M1V_TRACE_)
#include <stdio.h>
#endif
#include <string.h>

#include "MPEG2_SYS.h"
#include "SEQ_HEADER.h"

static const s8 default_intra_q[64] = { 8, 16, 19, 22, 26, 27, 29, 34,
                                       16, 16, 22, 24, 27, 29, 34, 37,
                                       19, 22, 26, 27, 29, 34, 34, 38,
                                       22, 22, 26, 27, 29, 34, 37, 40,
                                       22, 26, 27, 29, 32, 35, 40, 48,
                                       26, 27, 29, 32, 35, 40, 48, 58,
                                       26, 27, 29, 34, 38, 46, 56, 69,
                                       27, 29, 35, 38, 46, 56, 69, 83};

#define LOAD_INTRA(n, i1, i2) \
   case n+8: \
      hdr->intra_q_matrix[i1] |= *ptr >> 1; \
      hdr->intra_q_matrix[i2]  = *ptr << 7; \
      ptr++; \
      if (!(--data_len)) { \
         *cur_pos = n+9; \
         break; \
      }
#define LOAD_NON_INTRA(n, i1) \
   case n+72: \
      hdr->non_intra_q_matrix[i1] = *ptr; \
      ptr++; \
      if (!(--data_len)) { \
         *cur_pos = n+73; \
         break; \
      }

#ifdef _M2V_TRACE_
static void display_SEQ_HEADER(SEQ_HEADER *hdr);
#endif
#ifdef _M1V_TRACE_
static void display_SEQ_HEADER1(SEQ_HEADER *hdr);
#endif

/* Reads the sequence header in data_stream.
      INPUT: data_stream - Points to the data stream holding the sequence
                             header. This stream does not have to contain
                             the entire header. Repeated calls to this
                             function can be used to read the header in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the SEQ_HEADER that will hold the data.
              cur_pos - Used by MPEG2_VIDEO_MODULE.
              code    - Used by MPEG2_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 SEQ_HEADER_read(const u8         *data_stream,
                          u32         data_len,
                          SEQ_HEADER *hdr,
                          u32        *cur_pos,
                          u16        *code)
{
   const u8 *ptr = data_stream;

   switch (*cur_pos) {
   case 0:
      hdr->horiz_size  = (u16)(*ptr) << 4;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 1;
         break;
      }
   case 1:
/*      hdr->horiz_size |= (u16)(*ptr >> 4);*/
      if (*ptr >= 0x10)
         hdr->horiz_size += 0x0010;
      hdr->vert_size   = (u16)(*ptr & 0x0F) << 8;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 2;
         break;
      }
   case 2:
/*      hdr->vert_size  |= (u16)(*ptr);*/
      hdr->vert_size  |= (u16)(*ptr) & 0x00F0;
      if (*ptr & 0x0F)
         hdr->vert_size += 0x0010;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 3;
         break;
      }
   case 3:
      hdr->aspect_frame = *ptr;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 4;
         break;
      }
   case 4:
      hdr->bit_rate  = (u32)(*ptr) << 10;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 5;
         break;
      }
   case 5:
      hdr->bit_rate |= (u32)(*ptr) <<  2;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 6;
         break;
      }
   case 6:
      hdr->bit_rate |= (u32)(*ptr >> 6);
      hdr->vbv_buffer_size  = (u32)(*ptr & 0x1F) << 5;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 7;
         break;
      }
   case 7:
      hdr->vbv_buffer_size |= (u32)(*ptr >> 3);
      if (!(*ptr & 0x02)) {
         memcpy(hdr->intra_q_matrix, default_intra_q, 64);
         hdr->flags = 0x00;
         goto SEQ_HEADER_read_LOAD_NON_INTRA_Q_MATRIX;
      }

      hdr->flags             = 0x80;
      hdr->intra_q_matrix[0] = *ptr << 7;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 8;
         break;
      }
   LOAD_INTRA( 0,  0,  1)
   LOAD_INTRA( 1,  1,  8)
   LOAD_INTRA( 2,  8, 16)
   LOAD_INTRA( 3, 16,  9)
   LOAD_INTRA( 4,  9,  2)
   LOAD_INTRA( 5,  2,  3)
   LOAD_INTRA( 6,  3, 10)
   LOAD_INTRA( 7, 10, 17)
   LOAD_INTRA( 8, 17, 24)
   LOAD_INTRA( 9, 24, 32)
   LOAD_INTRA(10, 32, 25)
   LOAD_INTRA(11, 25, 18)
   LOAD_INTRA(12, 18, 11)
   LOAD_INTRA(13, 11,  4)
   LOAD_INTRA(14,  4,  5)
   LOAD_INTRA(15,  5, 12)
   LOAD_INTRA(16, 12, 19)
   LOAD_INTRA(17, 19, 26)
   LOAD_INTRA(18, 26, 33)
   LOAD_INTRA(19, 33, 40)
   LOAD_INTRA(20, 40, 48)
   LOAD_INTRA(21, 48, 41)
   LOAD_INTRA(22, 41, 34)
   LOAD_INTRA(23, 34, 27)
   LOAD_INTRA(24, 27, 20)
   LOAD_INTRA(25, 20, 13)
   LOAD_INTRA(26, 13,  6)
   LOAD_INTRA(27,  6,  7)
   LOAD_INTRA(28,  7, 14)
   LOAD_INTRA(29, 14, 21)
   LOAD_INTRA(30, 21, 28)
   LOAD_INTRA(31, 28, 35)
   LOAD_INTRA(32, 35, 42)
   LOAD_INTRA(33, 42, 49)
   LOAD_INTRA(34, 49, 56)
   LOAD_INTRA(35, 56, 57)
   LOAD_INTRA(36, 57, 50)
   LOAD_INTRA(37, 50, 43)
   LOAD_INTRA(38, 43, 36)
   LOAD_INTRA(39, 36, 29)
   LOAD_INTRA(40, 29, 22)
   LOAD_INTRA(41, 22, 15)
   LOAD_INTRA(42, 15, 23)
   LOAD_INTRA(43, 23, 30)
   LOAD_INTRA(44, 30, 37)
   LOAD_INTRA(45, 37, 44)
   LOAD_INTRA(46, 44, 51)
   LOAD_INTRA(47, 51, 58)
   LOAD_INTRA(48, 58, 59)
   LOAD_INTRA(49, 59, 52)
   LOAD_INTRA(50, 52, 45)
   LOAD_INTRA(51, 45, 38)
   LOAD_INTRA(52, 38, 31)
   LOAD_INTRA(53, 31, 39)
   LOAD_INTRA(54, 39, 46)
   LOAD_INTRA(55, 46, 53)
   LOAD_INTRA(56, 53, 60)
   LOAD_INTRA(57, 60, 61)
   LOAD_INTRA(58, 61, 54)
   LOAD_INTRA(59, 54, 47)
   LOAD_INTRA(60, 47, 55)
   LOAD_INTRA(61, 55, 62)
   LOAD_INTRA(62, 62, 63)
   case 71:
      hdr->intra_q_matrix[63] |= *ptr >> 1;
      
SEQ_HEADER_read_LOAD_NON_INTRA_Q_MATRIX:
      if (!(*ptr & 0x01)) {
         memset(hdr->non_intra_q_matrix, 16, 64);
         
         /* FINISHED READING THE SEQUENCE HEADER */
         ptr++;
         *cur_pos = 0;
         break;
      }
      ptr++;
      
      hdr->flags |= 0x40;
      if (!(--data_len)) {
         *cur_pos = 72;
          break;
      }
   LOAD_NON_INTRA( 0,  0)
   LOAD_NON_INTRA( 1,  1)
   LOAD_NON_INTRA( 2,  8)
   LOAD_NON_INTRA( 3, 16)
   LOAD_NON_INTRA( 4,  9)
   LOAD_NON_INTRA( 5,  2)
   LOAD_NON_INTRA( 6,  3)
   LOAD_NON_INTRA( 7, 10)
   LOAD_NON_INTRA( 8, 17)
   LOAD_NON_INTRA( 9, 24)
   LOAD_NON_INTRA(10, 32)
   LOAD_NON_INTRA(11, 25)
   LOAD_NON_INTRA(12, 18)
   LOAD_NON_INTRA(13, 11)
   LOAD_NON_INTRA(14,  4)
   LOAD_NON_INTRA(15,  5)
   LOAD_NON_INTRA(16, 12)
   LOAD_NON_INTRA(17, 19)
   LOAD_NON_INTRA(18, 26)
   LOAD_NON_INTRA(19, 33)
   LOAD_NON_INTRA(20, 40)
   LOAD_NON_INTRA(21, 48)
   LOAD_NON_INTRA(22, 41)
   LOAD_NON_INTRA(23, 34)
   LOAD_NON_INTRA(24, 27)
   LOAD_NON_INTRA(25, 20)
   LOAD_NON_INTRA(26, 13)
   LOAD_NON_INTRA(27,  6)
   LOAD_NON_INTRA(28,  7)
   LOAD_NON_INTRA(29, 14)
   LOAD_NON_INTRA(30, 21)
   LOAD_NON_INTRA(31, 28)
   LOAD_NON_INTRA(32, 35)
   LOAD_NON_INTRA(33, 42)
   LOAD_NON_INTRA(34, 49)
   LOAD_NON_INTRA(35, 56)
   LOAD_NON_INTRA(36, 57)
   LOAD_NON_INTRA(37, 50)
   LOAD_NON_INTRA(38, 43)
   LOAD_NON_INTRA(39, 36)
   LOAD_NON_INTRA(40, 29)
   LOAD_NON_INTRA(41, 22)
   LOAD_NON_INTRA(42, 15)
   LOAD_NON_INTRA(43, 23)
   LOAD_NON_INTRA(44, 30)
   LOAD_NON_INTRA(45, 37)
   LOAD_NON_INTRA(46, 44)
   LOAD_NON_INTRA(47, 51)
   LOAD_NON_INTRA(48, 58)
   LOAD_NON_INTRA(49, 59)
   LOAD_NON_INTRA(50, 52)
   LOAD_NON_INTRA(51, 45)
   LOAD_NON_INTRA(52, 38)
   LOAD_NON_INTRA(53, 31)
   LOAD_NON_INTRA(54, 39)
   LOAD_NON_INTRA(55, 46)
   LOAD_NON_INTRA(56, 53)
   LOAD_NON_INTRA(57, 60)
   LOAD_NON_INTRA(58, 61)
   LOAD_NON_INTRA(59, 54)
   LOAD_NON_INTRA(60, 47)
   LOAD_NON_INTRA(61, 55)
   LOAD_NON_INTRA(62, 62)
   case 135:
      hdr->non_intra_q_matrix[63] = *ptr;
      ptr++;
      /* FINISHED READING THE SEQUENCE HEADER */
      *cur_pos = 0;
      break;
   default:
#if defined(_M2V_TRACE_) || defined(_M1V_TRACE_)
      printf("SEQ_HEADER_read: ERROR illegal pos %u\n", *cur_pos);
#endif
      goto SEQ_HEADER_read_ERROR;
   }

   if (!(*cur_pos)) {
      *code |= 0x8000;     /* have to look for next start code */
      
      hdr->header_size = 8;
      if (SEQ_HEADER_INTRA_MATRIX_FLAG(hdr))
         hdr->header_size += 64;
      if (SEQ_HEADER_NON_INTRA_MATRIX_FLAG(hdr))
         hdr->header_size += 64;
#ifdef _M1V_TRACE_
      display_SEQ_HEADER1(hdr);
#endif
   }
   return (u32)(ptr - data_stream);
   
SEQ_HEADER_read_ERROR:
   *cur_pos = 0;
   *code    = 0x0FFF;
   return 0;
}

/* Reads the sequence header extension in data_stream.
      INPUT: data_stream - Points to the data stream holding the sequence
                             header extension. This stream does not have to
                             contain the entire extension. Repeated calls to
                             this function can be used to read the extension
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the SEQ_HEADER that will hold the data.
              cur_pos - Used by MPEG2_VIDEO_MODULE.
              code    - Used by MPEG2_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 SEQ_HEADER_EXT_read(const u8         *data_stream,
                              u32         data_len,
                              SEQ_HEADER *hdr,
                              u32        *cur_pos,
                              u16        *code)
{
   const u8 *ptr = data_stream;

   switch (*cur_pos) {
   case 0:
      hdr->profile_level  = *ptr << 4;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 1;
         break;
      }
   case 1:
      hdr->profile_level |= *ptr >> 4;
      hdr->flags         |= (*ptr & 0x06) >> 1;
      if (*ptr & 0x08)
         hdr->flags |= 0x20;
      if (*ptr & 0x01)
         hdr->horiz_size |= 0x2000;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 2;
         break;
      }
   case 2:
      if (*ptr & 0x80)
         hdr->horiz_size |= 0x1000;
      hdr->vert_size |= (u16)(*ptr & 0x60) <<  7;
      hdr->bit_rate  |= (u32)(*ptr & 0x1F) << 25;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 3;
         break;
      }
   case 3:
      hdr->bit_rate  |= (u32)(*ptr & 0xFE) << 17;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 4;
         break;
      }
   case 4:
      hdr->vbv_buffer_size |= (u32)(*ptr) << 10;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 5;
         break;
      }
   case 5:
      if (*ptr & 0x80)
         hdr->flags |= 0x10;
      hdr->frame_rate_ext = *ptr & 0x7F;
      ptr++;

      switch (SEQ_HEADER_FRAME_RATE_CODE(hdr)) {
      case FRAME_RATE_23976:
         hdr->frame_interval = (s32)((100.1*MPEG2_SYS_CLK_FREQ) / 2400.0);
         break;
      case FRAME_RATE_24:
         hdr->frame_interval = MPEG2_SYS_CLK_FREQ / 24;
         break;
      case FRAME_RATE_25:
      case FRAME_RATE_50:
         hdr->frame_interval = MPEG2_SYS_CLK_FREQ / 25;
         break;
      case FRAME_RATE_2997:
      case FRAME_RATE_5994:
         hdr->frame_interval = (s32)((100.1*MPEG2_SYS_CLK_FREQ) / 3000.0);
         break;
      case FRAME_RATE_30:
      case FRAME_RATE_60:
      default:
         hdr->frame_interval = MPEG2_SYS_CLK_FREQ / 30;
      }
      if (hdr->frame_rate_ext) {
         hdr->frame_interval *= (u32)(SEQ_HEADER_FRAME_RATE_EXT_D(hdr) + 1);
         hdr->frame_interval /= (u32)(SEQ_HEADER_FRAME_RATE_EXT_N(hdr) + 1);
      }

      /* FINISHED READING THE SEQUENCE HEADER EXTENSION */
      *cur_pos      = 0;
      *code        |= 0x8000;    /* have to look for next start code */
      hdr->ext_size = 6;
#ifdef _M2V_TRACE_
      display_SEQ_HEADER(hdr);
#endif
      break;
   default:
#ifdef _M2V_TRACE_
      printf("SEQ_HEADER_EXT_read: ERROR illegal pos %u\n", *cur_pos);
#endif
      *cur_pos = 0;
      *code    = 0x0FFF;
      return 0;
   }

   return (u32)(ptr - data_stream);
}


#undef LOAD_INTRA
#undef LOAD_NON_INTRA
#define LOAD_INTRA(n, i1, i2) \
   case n+1: \
      hdr->intra_q_matrix[i1] |= *ptr >> 3; \
      hdr->intra_q_matrix[i2]  = *ptr << 5; \
      ptr++; \
      if (!(--data_len)) { \
         *cur_pos = n+2; \
         break; \
      }
#define LOAD_NON_INTRA(n, i1, i2) \
   case n+65: \
      hdr->non_intra_q_matrix[i1] |= *ptr >> 2; \
      hdr->non_intra_q_matrix[i2]  = *ptr << 6; \
      ptr++; \
      if (!(--data_len)) { \
         *cur_pos = n+66; \
         break; \
      }
/* Reads the quant matrix extension in data_stream.
      INPUT: data_stream - Points to the data stream holding the quant
                             matrix extension. This stream does not have to
                             contain the entire extension. Repeated calls to
                             this function can be used to read the extension
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the SEQ_HEADER that will hold the data.
              cur_pos - Used by MPEG2_VIDEO_MODULE.
              code    - Used by MPEG2_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 QUANT_MATRIX_EXT_read(const u8         *data_stream,
                                u32         data_len,
                                SEQ_HEADER *hdr,
                                u32        *cur_pos,
                                u16        *code)
{
   const u8 *ptr = data_stream;

   switch (*cur_pos) {
   case 0:
      if (!(*ptr & 0x08))
         goto QUANT_MATRIX_read_LOAD_NON_INTRA_Q_MATRIX;

      hdr->intra_q_matrix[0] = *ptr << 5;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 1;
         break;
      }
   LOAD_INTRA( 0,  0,  1)
   LOAD_INTRA( 1,  1,  8)
   LOAD_INTRA( 2,  8, 16)
   LOAD_INTRA( 3, 16,  9)
   LOAD_INTRA( 4,  9,  2)
   LOAD_INTRA( 5,  2,  3)
   LOAD_INTRA( 6,  3, 10)
   LOAD_INTRA( 7, 10, 17)
   LOAD_INTRA( 8, 17, 24)
   LOAD_INTRA( 9, 24, 32)
   LOAD_INTRA(10, 32, 25)
   LOAD_INTRA(11, 25, 18)
   LOAD_INTRA(12, 18, 11)
   LOAD_INTRA(13, 11,  4)
   LOAD_INTRA(14,  4,  5)
   LOAD_INTRA(15,  5, 12)
   LOAD_INTRA(16, 12, 19)
   LOAD_INTRA(17, 19, 26)
   LOAD_INTRA(18, 26, 33)
   LOAD_INTRA(19, 33, 40)
   LOAD_INTRA(20, 40, 48)
   LOAD_INTRA(21, 48, 41)
   LOAD_INTRA(22, 41, 34)
   LOAD_INTRA(23, 34, 27)
   LOAD_INTRA(24, 27, 20)
   LOAD_INTRA(25, 20, 13)
   LOAD_INTRA(26, 13,  6)
   LOAD_INTRA(27,  6,  7)
   LOAD_INTRA(28,  7, 14)
   LOAD_INTRA(29, 14, 21)
   LOAD_INTRA(30, 21, 28)
   LOAD_INTRA(31, 28, 35)
   LOAD_INTRA(32, 35, 42)
   LOAD_INTRA(33, 42, 49)
   LOAD_INTRA(34, 49, 56)
   LOAD_INTRA(35, 56, 57)
   LOAD_INTRA(36, 57, 50)
   LOAD_INTRA(37, 50, 43)
   LOAD_INTRA(38, 43, 36)
   LOAD_INTRA(39, 36, 29)
   LOAD_INTRA(40, 29, 22)
   LOAD_INTRA(41, 22, 15)
   LOAD_INTRA(42, 15, 23)
   LOAD_INTRA(43, 23, 30)
   LOAD_INTRA(44, 30, 37)
   LOAD_INTRA(45, 37, 44)
   LOAD_INTRA(46, 44, 51)
   LOAD_INTRA(47, 51, 58)
   LOAD_INTRA(48, 58, 59)
   LOAD_INTRA(49, 59, 52)
   LOAD_INTRA(50, 52, 45)
   LOAD_INTRA(51, 45, 38)
   LOAD_INTRA(52, 38, 31)
   LOAD_INTRA(53, 31, 39)
   LOAD_INTRA(54, 39, 46)
   LOAD_INTRA(55, 46, 53)
   LOAD_INTRA(56, 53, 60)
   LOAD_INTRA(57, 60, 61)
   LOAD_INTRA(58, 61, 54)
   LOAD_INTRA(59, 54, 47)
   LOAD_INTRA(60, 47, 55)
   LOAD_INTRA(61, 55, 62)
   LOAD_INTRA(62, 62, 63)
   case 64:
      hdr->intra_q_matrix[63] |= *ptr >> 3;

QUANT_MATRIX_read_LOAD_NON_INTRA_Q_MATRIX:
      if (!(*ptr & 0x04)) {
         /* FINISHED READING THE QUANT MATRIX EXT */
         ptr++;
         *cur_pos = 0;
         break;
      }

      hdr->non_intra_q_matrix[0] = *ptr << 6;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 65;
         break;
      }
   LOAD_NON_INTRA( 0,  0,  1)
   LOAD_NON_INTRA( 1,  1,  8)
   LOAD_NON_INTRA( 2,  8, 16)
   LOAD_NON_INTRA( 3, 16,  9)
   LOAD_NON_INTRA( 4,  9,  2)
   LOAD_NON_INTRA( 5,  2,  3)
   LOAD_NON_INTRA( 6,  3, 10)
   LOAD_NON_INTRA( 7, 10, 17)
   LOAD_NON_INTRA( 8, 17, 24)
   LOAD_NON_INTRA( 9, 24, 32)
   LOAD_NON_INTRA(10, 32, 25)
   LOAD_NON_INTRA(11, 25, 18)
   LOAD_NON_INTRA(12, 18, 11)
   LOAD_NON_INTRA(13, 11,  4)
   LOAD_NON_INTRA(14,  4,  5)
   LOAD_NON_INTRA(15,  5, 12)
   LOAD_NON_INTRA(16, 12, 19)
   LOAD_NON_INTRA(17, 19, 26)
   LOAD_NON_INTRA(18, 26, 33)
   LOAD_NON_INTRA(19, 33, 40)
   LOAD_NON_INTRA(20, 40, 48)
   LOAD_NON_INTRA(21, 48, 41)
   LOAD_NON_INTRA(22, 41, 34)
   LOAD_NON_INTRA(23, 34, 27)
   LOAD_NON_INTRA(24, 27, 20)
   LOAD_NON_INTRA(25, 20, 13)
   LOAD_NON_INTRA(26, 13,  6)
   LOAD_NON_INTRA(27,  6,  7)
   LOAD_NON_INTRA(28,  7, 14)
   LOAD_NON_INTRA(29, 14, 21)
   LOAD_NON_INTRA(30, 21, 28)
   LOAD_NON_INTRA(31, 28, 35)
   LOAD_NON_INTRA(32, 35, 42)
   LOAD_NON_INTRA(33, 42, 49)
   LOAD_NON_INTRA(34, 49, 56)
   LOAD_NON_INTRA(35, 56, 57)
   LOAD_NON_INTRA(36, 57, 50)
   LOAD_NON_INTRA(37, 50, 43)
   LOAD_NON_INTRA(38, 43, 36)
   LOAD_NON_INTRA(39, 36, 29)
   LOAD_NON_INTRA(40, 29, 22)
   LOAD_NON_INTRA(41, 22, 15)
   LOAD_NON_INTRA(42, 15, 23)
   LOAD_NON_INTRA(43, 23, 30)
   LOAD_NON_INTRA(44, 30, 37)
   LOAD_NON_INTRA(45, 37, 44)
   LOAD_NON_INTRA(46, 44, 51)
   LOAD_NON_INTRA(47, 51, 58)
   LOAD_NON_INTRA(48, 58, 59)
   LOAD_NON_INTRA(49, 59, 52)
   LOAD_NON_INTRA(50, 52, 45)
   LOAD_NON_INTRA(51, 45, 38)
   LOAD_NON_INTRA(52, 38, 31)
   LOAD_NON_INTRA(53, 31, 39)
   LOAD_NON_INTRA(54, 39, 46)
   LOAD_NON_INTRA(55, 46, 53)
   LOAD_NON_INTRA(56, 53, 60)
   LOAD_NON_INTRA(57, 60, 61)
   LOAD_NON_INTRA(58, 61, 54)
   LOAD_NON_INTRA(59, 54, 47)
   LOAD_NON_INTRA(60, 47, 55)
   LOAD_NON_INTRA(61, 55, 62)
   LOAD_NON_INTRA(62, 62, 63)
   case 128:
      hdr->non_intra_q_matrix[63] |= *ptr >> 2;
      ptr++;
      /* FINISHED READING THE QUANT MATRIX EXT */
      *cur_pos = 0;
      break;
   default:
#ifdef _M2V_TRACE_
      printf("QUANT_MATRIX_read: ERROR illegal pos %u\n", *cur_pos);
#endif
      goto QUANT_MATRIX_read_ERROR;
   }

   if (!(*cur_pos))
      *code |= 0x8000;     /* have to look for next start code */
   return (u32)(ptr - data_stream);

QUANT_MATRIX_read_ERROR:
   *cur_pos = 0;
   *code    = 0x0FFF;
   return 0;
}

#ifdef _M2V_TRACE_
static void display_SEQ_HEADER(SEQ_HEADER *hdr)
{
   const
   char  *aspect_ratio, *profile_ID, *level_ID, *chroma_format;
   double frame_rate;
   int    i;

   switch (SEQ_HEADER_ASPECT_RATIO(hdr)) {
   case ASPECT_RATIO_FORBIDDEN:
      aspect_ratio = "FORBIDDEN";
      break;
   case ASPECT_RATIO_SAR_11:
      aspect_ratio = "SAR = 1:1";
      break;
   case ASPECT_RATIO_DAR_34:
      aspect_ratio = "DAR = 3:4";
      break;
   case ASPECT_RATIO_DAR_916:
      aspect_ratio = "DAR = 9:16";
      break;
   case ASPECT_RATIO_DAR_1221:
      aspect_ratio = "DAR = 1:2.21";
      break;
   default:
      aspect_ratio = "RESERVED";
   };

   switch (SEQ_HEADER_PROFILE_ID(hdr)) {
   case PROFILE_HIGH:
      profile_ID = "High";
      break;
   case PROFILE_SPATIALLY_SCALABLE:
      profile_ID = "Spatially Scalable";
      break;
   case PROFILE_SNR_SCALABLE:
      profile_ID = "SNR Scalable";
      break;
   case PROFILE_MAIN:
      profile_ID = "Main";
      break;
   case PROFILE_SIMPLE:
      profile_ID = "Simple";
      break;
   default:
      profile_ID = "RESERVED";
   };

   switch (SEQ_HEADER_LEVEL_ID(hdr)) {
   case LEVEL_HIGH:
      level_ID = "High";
      break;
   case LEVEL_HIGH_1440:
      level_ID = "High 1440";
      break;
   case LEVEL_MAIN:
      level_ID = "Main";
      break;
   case LEVEL_LOW:
      level_ID = "Low";
      break;
   default:
      level_ID = "RESERVED";
   };
   
   switch (SEQ_HEADER_FRAME_RATE_CODE(hdr)) {
   case FRAME_RATE_FORBIDDEN:
      frame_rate = 0.0;
      break;
   case FRAME_RATE_23976:
      frame_rate = 24000.0 / 1001.0;
      break;
   case FRAME_RATE_24:
      frame_rate = 24.0;
      break;
   case FRAME_RATE_25:
      frame_rate = 25.0;
      break;
   case FRAME_RATE_2997:
      frame_rate = 30000.0 / 1001.0;
      break;
   case FRAME_RATE_30:
      frame_rate = 30.0;
      break;
   case FRAME_RATE_50:
      frame_rate = 50.0;
      break;
   case FRAME_RATE_5994:
      frame_rate = 60000.0 / 1001.0;
      break;
   case FRAME_RATE_60:
      frame_rate = 60.0;
      break;
   default:
      frame_rate = -1.0;
   };
   if (hdr->frame_rate_ext) {
      frame_rate *= (double)(SEQ_HEADER_FRAME_RATE_EXT_N(hdr) + 1) /
                    (double)(SEQ_HEADER_FRAME_RATE_EXT_D(hdr) + 1);
   }

   switch (SEQ_HEADER_CHROMA_FORMAT(hdr)) {
   case CHROMA_FORMAT_RESERVED:
      chroma_format = "RESERVED";
      break;
   case CHROMA_FORMAT_420:
      chroma_format = "4:2:0 (4Y, 1Cb, 1Cr)";
      break;
   case CHROMA_FORMAT_422:
      chroma_format = "4:2:2 (4Y, 2Cb, 2Cr)";
      break;
   case CHROMA_FORMAT_444:
      chroma_format = "4:4:4 (4Y, 4Cb, 4Cr)";
   };
   
   printf("SEQ_HEADER> profile / level = %s / %s\n"
          "            chroma_format   = %s\n"
          "            aspect_ratio    = %s\n"
          "            frame_size      = %hux%hu\n"
          "            frame_rate      = %gfps (%uclks)\n"
          "            bit_rate        = %gMbps\n"
          "            progressive     = %c   low_delay = %c\n"
          "            vbv_buffer_size = %ukB\n",
          profile_ID, level_ID,
          chroma_format,
          aspect_ratio,
          hdr->horiz_size, hdr->vert_size,
          frame_rate, hdr->frame_interval,
          (double)hdr->bit_rate * 0.0004,
          SEQ_HEADER_PROGRESSIVE_FLAG(hdr) ? 'Y' : 'N',
          SEQ_HEADER_LOW_DELAY_FLAG(hdr) ? 'Y' : 'N',
          hdr->vbv_buffer_size * 2);

   printf(" intra_q_matrix");
   if (SEQ_HEADER_INTRA_MATRIX_FLAG(hdr))
      printf("*");
   else
      printf(" ");
   printf("                   non_intra_q_matrix");
   if (SEQ_HEADER_NON_INTRA_MATRIX_FLAG(hdr))
      printf("*\n");
   else
      printf(" \n");

      for (i = 0; i < 8; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 0; i < 8; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 16; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 8; i < 16; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 24; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 16; i < 24; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");

      for ( ; i < 32; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 24; i < 32; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 40; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 32; i < 40; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 48; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 40; i < 48; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 56; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 48; i < 56; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 64; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 56; i < 64; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
}
#endif

#ifdef _M1V_TRACE_
static void display_SEQ_HEADER1(SEQ_HEADER *hdr)
{
   double frame_rate, aspect_ratio;
   int    i;

   switch (SEQ_HEADER1_ASPECT_RATIO(hdr)) {
   case ASPECT_RATIO_FORBIDDEN:
      aspect_ratio =  0.0;
      break;
   case ASPECT_RATIO_SAR_10:
      aspect_ratio =  1.0;
      break;
   case ASPECT_RATIO_SAR_06735:
      aspect_ratio =  0.6735;
      break;
   case ASPECT_RATIO_SAR_07031:
      aspect_ratio =  0.7031;
      break;
   case ASPECT_RATIO_SAR_07615:
      aspect_ratio =  0.7615;
      break;
   case ASPECT_RATIO_SAR_08055:
      aspect_ratio =  0.8055;
      break;
   case ASPECT_RATIO_SAR_08437:
      aspect_ratio =  0.8437;
      break;
   case ASPECT_RATIO_SAR_08935:
      aspect_ratio =  0.8935;
      break;
   case ASPECT_RATIO_SAR_09157:
      aspect_ratio =  0.9157;
      break;
   case ASPECT_RATIO_SAR_09815:
      aspect_ratio =  0.9815;
      break;
   case ASPECT_RATIO_SAR_10255:
      aspect_ratio =  1.0255;
      break;
   case ASPECT_RATIO_SAR_10695:
      aspect_ratio =  1.0695;
      break;
   case ASPECT_RATIO_SAR_10950:
      aspect_ratio =  1.095;
      break;
   case ASPECT_RATIO_SAR_11575:
      aspect_ratio =  1.1575;
      break;
   case ASPECT_RATIO_SAR_12015:
      aspect_ratio =  1.2015;
      break;
   case ASPECT_RATIO_RESERVED:
      aspect_ratio = -1.0;
   };

   switch (SEQ_HEADER_FRAME_RATE_CODE(hdr)) {
   case FRAME_RATE_FORBIDDEN:
      frame_rate = 0.0;
      break;
   case FRAME_RATE_23976:
      frame_rate = 24000.0 / 1001.0;
      break;
   case FRAME_RATE_24:
      frame_rate = 24.0;
      break;
   case FRAME_RATE_25:
      frame_rate = 25.0;
      break;
   case FRAME_RATE_2997:
      frame_rate = 30000.0 / 1001.0;
      break;
   case FRAME_RATE_30:
      frame_rate = 30.0;
      break;
   case FRAME_RATE_50:
      frame_rate = 50.0;
      break;
   case FRAME_RATE_5994:
      frame_rate = 60000.0 / 1001.0;
      break;
   case FRAME_RATE_60:
      frame_rate = 60.0;
      break;
   default:
      frame_rate = -1.0;
   };

   printf("SEQ_HEADER1> aspect_ratio    = %.4lf\n"
          "             frame_size      = %hux%hu\n"
          "             frame_rate      = %gfps (%uclks)\n"
          "             bit_rate        = %gMbps\n"
          "             vbv_buffer_size = %ukB\n",
          aspect_ratio,
          hdr->horiz_size, hdr->vert_size,
          frame_rate, hdr->frame_interval,
          (double)hdr->bit_rate * 0.0004,
          hdr->vbv_buffer_size * 2);

   printf(" intra_q_matrix");
   if (SEQ_HEADER_INTRA_MATRIX_FLAG(hdr))
      printf("*");
   else
      printf(" ");
   printf("                   non_intra_q_matrix");
   if (SEQ_HEADER_NON_INTRA_MATRIX_FLAG(hdr))
      printf("*\n");
   else
      printf(" \n");

      for (i = 0; i < 8; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 0; i < 8; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 16; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 8; i < 16; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 24; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 16; i < 24; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");

      for ( ; i < 32; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 24; i < 32; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 40; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 32; i < 40; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 48; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 40; i < 48; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 56; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 48; i < 56; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
      
      for ( ; i < 64; i++) {
         printf(" %3hu", (u16)hdr->intra_q_matrix[i]);
      }
      printf("  ");
      for (i = 56; i < 64; i++) {
         printf(" %3hu", (u16)hdr->non_intra_q_matrix[i]);
      }
      printf("\n");
}
#endif

#undef LOAD_INTRA
#undef LOAD_NON_INTRA

