/***************************************************************************
                          idct.c  -  description
                             -------------------
   IDCT function.
                             -------------------
    begin                : Sun Nov 14 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "idct.h"


#if defined(__MMX__)

# include "mmx/mmx_idct.c"

#else
/******************************************************************************/
/* This is the default idct function, that will work on all systems.
     Operations per row  :  22 mults,  32 adds,   8 shifts
     Operations per col  :  22 mults,  31 adds,   8 shifts
     Operations per block: 352 mults, 504 adds, 128 shifts */

#define shift_value  14
#define row_shift    (shift_value-3)
#define row_error    (1 << (row_shift-1))
#define col_shift    (shift_value+6)
#define col_error    ((1 << (col_shift-1)) / cosp4_16)

/* cospk_16r = (s32)( ( cos(PI*k / 16) * sqrt(2) * (1 << shift_value)) + 0.5 ) */
#define cosp1_16  22725
#define cosp2_16  21407
#define cosp3_16  19266
#define cosp4_16  16384
#define cosp5_16  12873
#define cosp6_16   8867
#define cosp7_16   4520

/* cosmk_16 = (s32)( (-cos(PI*k / 16) * sqrt(2) * (1 << shift_value)) + 0.5 ) */
#define cosm1_16 -22724
#define cosm2_16 -21406
#define cosm3_16 -19265
#define cosm4_16 -16384
#define cosm5_16 -12872
#define cosm6_16  -8866
#define cosm7_16  -4519

/* Does an 8x8 2D iDCT operation.
      INPUT: data - An array holding the 64 dct coefficients.
      OUTPUT: data - The 64 dct coefficients are replaced by the 64
                       recovered samples. */
#ifdef _IEEE_TEST_
void j_rev_dct(DCTBLOCK data)
#else
void idct(s16 *data)
#endif
{
   s32  t0, t1, t2, t3, t4, t5, t6, t7;
   s16 *end_data = data;

   /* process rows */
   data += 8*8;
   do {
      data -= 8;

      t7 =  cosp4_16 * (s32)(data[0] + data[4]);
      t6 =  cosp4_16 * (s32)(data[0] - data[4]);
      t5 =  cosp2_16 * (s32)data[2];
      t4 =  cosp6_16 * (s32)data[2];
      t3 = -cosp6_16 * (s32)data[6];
      t2 =  cosp2_16 * (s32)data[6];
      
      /* these adds are needed to improve accuracy */
      t7 += row_error;
      t6 += row_error;

      t0  = t7 + t5 - t3;
      t1  = t6 + t4 - t2;
      t2 += t6 - t4;
      t3 += t7 - t5;

      t4 = (cosp1_16 * (s32)data[1]) + (cosp3_16 * (s32)data[3]) +
           (cosp5_16 * (s32)data[5]) + (cosp7_16 * (s32)data[7]);
      t5 = (cosp3_16 * (s32)data[1]) + (cosm7_16 * (s32)data[3]) +
           (cosm1_16 * (s32)data[5]) + (cosm5_16 * (s32)data[7]);
      t6 = (cosp5_16 * (s32)data[1]) + (cosm1_16 * (s32)data[3]) +
           (cosp7_16 * (s32)data[5]) + (cosp3_16 * (s32)data[7]);
      t7 = (cosp7_16 * (s32)data[1]) + (cosm5_16 * (s32)data[3]) +
           (cosp3_16 * (s32)data[5]) + (cosm1_16 * (s32)data[7]);

      data[0] = (s16)((t0 + t4) >> row_shift);
      data[7] = (s16)((t0 - t4) >> row_shift);
      data[1] = (s16)((t1 + t5) >> row_shift);
      data[6] = (s16)((t1 - t5) >> row_shift);
      data[2] = (s16)((t2 + t6) >> row_shift);
      data[5] = (s16)((t2 - t6) >> row_shift);
      data[3] = (s16)((t3 + t7) >> row_shift);
      data[4] = (s16)((t3 - t7) >> row_shift);
   } while (data != end_data);

   /* process cols */
   end_data += 8;
   do {
      data[8*0] += col_error; /* this add is needed to improve accuracy */

      t7 =  cosp4_16 * (s32)(data[8*0] + data[8*4]);
      t6 =  cosp4_16 * (s32)(data[8*0] - data[8*4]);
      t5 =  cosp2_16 * (s32)data[8*2];
      t4 =  cosp6_16 * (s32)data[8*2];
      t3 = -cosp6_16 * (s32)data[8*6];
      t2 =  cosp2_16 * (s32)data[8*6];

      t0  = t7 + t5 - t3;
      t1  = t6 + t4 - t2;
      t2 += t6 - t4;
      t3 += t7 - t5;

      t4 = (cosp1_16 * (s32)data[8*1]) + (cosp3_16 * (s32)data[8*3]) +
           (cosp5_16 * (s32)data[8*5]) + (cosp7_16 * (s32)data[8*7]);
      t5 = (cosp3_16 * (s32)data[8*1]) + (cosm7_16 * (s32)data[8*3]) +
           (cosm1_16 * (s32)data[8*5]) + (cosm5_16 * (s32)data[8*7]);
      t6 = (cosp5_16 * (s32)data[8*1]) + (cosm1_16 * (s32)data[8*3]) +
           (cosp7_16 * (s32)data[8*5]) + (cosp3_16 * (s32)data[8*7]);
      t7 = (cosp7_16 * (s32)data[8*1]) + (cosm5_16 * (s32)data[8*3]) +
           (cosp3_16 * (s32)data[8*5]) + (cosm1_16 * (s32)data[8*7]);

      data[8*0] = (s16)((t0 + t4) >> col_shift);
      data[8*7] = (s16)((t0 - t4) >> col_shift);
      data[8*1] = (s16)((t1 + t5) >> col_shift);
      data[8*6] = (s16)((t1 - t5) >> col_shift);
      data[8*2] = (s16)((t2 + t6) >> col_shift);
      data[8*5] = (s16)((t2 - t6) >> col_shift);
      data[8*3] = (s16)((t3 + t7) >> col_shift);
      data[8*4] = (s16)((t3 - t7) >> col_shift);

      data++;
   } while (data != end_data);
}

/*
IEEE test conditions: -L = -256, +H = 255, sign = 1, #iters = 10000
Peak absolute values of errors:
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
Worst peak error = 1  (meets spec limit 1)

Mean square errors:
   0.0064   0.0082   0.0073   0.0062   0.0051   0.0064   0.0081   0.0067
   0.0073   0.0073   0.0075   0.0068   0.0068   0.0089   0.0069   0.0065
   0.0059   0.0077   0.0059   0.0066   0.0073   0.0082   0.0069   0.0062
   0.0079   0.0075   0.0069   0.0080   0.0077   0.0073   0.0083   0.0074
   0.0066   0.0091   0.0081   0.0076   0.0074   0.0075   0.0091   0.0071
   0.0065   0.0065   0.0058   0.0074   0.0069   0.0064   0.0076   0.0064
   0.0076   0.0086   0.0084   0.0078   0.0072   0.0079   0.0068   0.0074
   0.0064   0.0075   0.0071   0.0062   0.0066   0.0073   0.0071   0.0059
Worst pmse = 0.009100  (meets spec limit 0.06)
Overall mse = 0.007186  (meets spec limit 0.02)

Mean errors:
  -0.0004  -0.0006  -0.0003   0.0006   0.0005   0.0010   0.0011   0.0001
  -0.0005  -0.0011  -0.0011   0.0000   0.0002  -0.0005  -0.0009   0.0005
  -0.0011  -0.0015  -0.0003  -0.0006  -0.0003   0.0004   0.0009  -0.0006
  -0.0013  -0.0005  -0.0011   0.0004   0.0003   0.0007  -0.0015  -0.0006
  -0.0002   0.0021   0.0001   0.0002   0.0016  -0.0005  -0.0021   0.0015
   0.0001  -0.0009   0.0000  -0.0006   0.0005  -0.0002   0.0006  -0.0012
   0.0004  -0.0004  -0.0004   0.0006  -0.0002  -0.0001   0.0000   0.0000
   0.0002   0.0011   0.0005   0.0000   0.0008  -0.0007  -0.0007  -0.0015
Worst mean error = 0.002100  (meets spec limit 0.015)
Overall mean error = -0.000117  (meets spec limit 0.0015)

0 elements of IDCT(0) were not zero

**************************************************************************

IEEE test conditions: -L = -5, +H = 5, sign = 1, #iters = 10000
Peak absolute values of errors:
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
Worst peak error = 1  (meets spec limit 1)

Mean square errors:
   0.0074   0.0067   0.0050   0.0063   0.0065   0.0059   0.0056   0.0057
   0.0065   0.0065   0.0060   0.0061   0.0063   0.0055   0.0087   0.0062
   0.0060   0.0069   0.0068   0.0057   0.0064   0.0072   0.0069   0.0067
   0.0058   0.0056   0.0055   0.0074   0.0080   0.0058   0.0067   0.0056
   0.0052   0.0068   0.0075   0.0069   0.0054   0.0069   0.0062   0.0065
   0.0074   0.0066   0.0059   0.0052   0.0063   0.0054   0.0064   0.0068
   0.0053   0.0047   0.0074   0.0058   0.0056   0.0074   0.0051   0.0067
   0.0073   0.0063   0.0058   0.0062   0.0062   0.0066   0.0067   0.0054
Worst pmse = 0.008700  (meets spec limit 0.06)
Overall mse = 0.006309  (meets spec limit 0.02)

Mean errors:
   0.0004   0.0007  -0.0006  -0.0003  -0.0001  -0.0007   0.0002   0.0011
   0.0009   0.0005   0.0002  -0.0017   0.0007  -0.0001   0.0007   0.0006
  -0.0004  -0.0001  -0.0002   0.0009   0.0000   0.0008   0.0007   0.0011
   0.0006  -0.0010  -0.0015  -0.0002  -0.0006   0.0000   0.0015  -0.0002
   0.0002   0.0002  -0.0005   0.0001   0.0006   0.0011   0.0002  -0.0011
   0.0000   0.0004   0.0013   0.0014   0.0017   0.0010  -0.0002  -0.0018
   0.0011   0.0003  -0.0006  -0.0002   0.0012  -0.0008  -0.0001   0.0007
   0.0007  -0.0001  -0.0010  -0.0002   0.0004  -0.0008   0.0003  -0.0002
Worst mean error = 0.001800  (meets spec limit 0.015)
Overall mean error = 0.000144  (meets spec limit 0.0015)

0 elements of IDCT(0) were not zero

**************************************************************************

IEEE test conditions: -L = -300, +H = 300, sign = 1, #iters = 10000
Peak absolute values of errors:
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
Worst peak error = 1  (meets spec limit 1)

Mean square errors:
   0.0062   0.0064   0.0073   0.0057   0.0055   0.0067   0.0081   0.0055
   0.0061   0.0059   0.0069   0.0050   0.0059   0.0060   0.0075   0.0058
   0.0066   0.0060   0.0077   0.0069   0.0058   0.0071   0.0071   0.0080
   0.0053   0.0066   0.0080   0.0055   0.0066   0.0059   0.0057   0.0055
   0.0061   0.0066   0.0057   0.0060   0.0064   0.0052   0.0082   0.0069
   0.0057   0.0058   0.0058   0.0055   0.0050   0.0067   0.0048   0.0056
   0.0078   0.0064   0.0070   0.0072   0.0066   0.0046   0.0067   0.0054
   0.0059   0.0062   0.0048   0.0065   0.0058   0.0055   0.0056   0.0059
Worst pmse = 0.008200  (meets spec limit 0.06)
Overall mse = 0.006230  (meets spec limit 0.02)

Mean errors:
  -0.0004   0.0000   0.0009   0.0009  -0.0017  -0.0001  -0.0011   0.0005
  -0.0003  -0.0001  -0.0001   0.0002  -0.0003  -0.0006   0.0015   0.0012
  -0.0004   0.0000   0.0007   0.0005   0.0002  -0.0007   0.0001   0.0010
   0.0001  -0.0006   0.0034   0.0001  -0.0008  -0.0003  -0.0007  -0.0007
   0.0003   0.0000   0.0011  -0.0008  -0.0004   0.0008   0.0010   0.0003
   0.0013   0.0008   0.0014  -0.0005  -0.0002  -0.0003  -0.0002  -0.0012
  -0.0014  -0.0010   0.0010  -0.0012  -0.0004  -0.0002  -0.0009  -0.0002
  -0.0005  -0.0016   0.0008  -0.0011   0.0002   0.0005   0.0012   0.0003
Worst mean error = 0.003400  (meets spec limit 0.015)
Overall mean error = 0.000020  (meets spec limit 0.0015)

0 elements of IDCT(0) were not zero
*/

#undef shift_value
#undef row_shift
#undef row_error
#undef col_shift
#undef col_error
#undef cosp1_16
#undef cosp2_16
#undef cosp3_16
#undef cosp4_16
#undef cosp5_16
#undef cosp6_16
#undef cosp7_16
#undef cosm1_16
#undef cosm2_16
#undef cosm3_16
#undef cosm4_16
#undef cosm5_16
#undef cosm6_16
#undef cosm7_16

#endif
