/***************************************************************************
                     VC1_video_module.c  -  description
                             -------------------
   VC1 video module functions.
                             -------------------
    begin                : Tue May 16 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define OUTPUT_BLOCK_TRACE*/
#if defined(_SHOW_ERRORS_)
# include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>

#include "MPEG2_SYS.h"
#include "PES_PKT.h"
#include "commands.h"
#include "display/display.h"
#include "mpeg2_video/motion_vectors.h"
#include "VC1_video_module.h"
#include "wmv.h"
#include "SEQ_LAYER.h"
#include "PIC_LAYER.h"
#include "I_MACROBLOCK.h"
#include "P_MACROBLOCK.h"


typedef struct {
   WMV_DATA   wmv;
   SEQ_LAYER  seq;
   PIC_LAYER  pic;
   SYS_DATA  *sys_data;
   CLK_VAL    offset, cur_pts, pes_pts, output_delay;
} VC1_video_module_data;


static void simple_main_decode(VC1_video_module_data *data);
static void init_display(VC1_video_module_data *data);
static void new_frame(VC1_video_module_data *data);


/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t VC1_VIDEO_MODULE_mem_needed()
{
   return sizeof(VC1_video_module_data);
}

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void VC1_VIDEO_MODULE_init(void *mem, SYS_DATA *sys_data)
{
   VC1_video_module_data *data = (VC1_video_module_data *)mem;

   data->wmv.state = WMV_next_guid_STATE;
   if (sys_data)
      data->sys_data = sys_data;
}

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void VC1_VIDEO_MODULE_done(void *mem)
{
   VC1_video_module_data *data = (VC1_video_module_data *)mem;

   if (data->pic.pred_frame) {
      if (data->sys_data->command == CMD_STOP) {
         DISPLAY_done();
         data->pic.pred_frame = NULL;
      }
   }
   data->wmv.state = WMV_next_guid_STATE;
}

/* Gives the module the offset value needed to obtain the program clock value.
      INPUT: void *  - Pointer to the memory provided for the module.
             CLK_VAL - offset value */
void VC1_VIDEO_MODULE_set_offset(void *mem, CLK_VAL offset)
{
   VC1_video_module_data *data = (VC1_video_module_data *)mem;

   if (data->sys_data->command != CMD_PAUSE)
      data->cur_pts += offset - data->offset;
   data->offset = offset;
}

/* Alerts the module to a discontinuity in the data stream.
      INPUT: void * - Pointer to the memory provided for the module. */
void VC1_VIDEO_MODULE_discontinuity(void *mem)
{
   VC1_video_module_data *data = (VC1_video_module_data *)mem;

   data->wmv.state = WMV_next_guid_STATE;
}


/**************************************************************************
   PES MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the  PES_PKTs
       they produce.
 **************************************************************************/
/* Initialize everything needed to handle the given elementary stream.
     The PES module will handle the elementary stream (if able) until the
     MPEG2_MODULE_done function is called, or this function is called
     again.
      INPUT: void *        - Pointer to the memory provided for the module.
             PMT_SECTION * - Pointer to the PMT data.
             int           - Index, in elem array, of the elementary stream
                               that needs to be handled.
      RETURN: Module is ready to handle
                the stream.               - Non-zero value.
              Module cannot handle the
                stream.                   - 0 */
int VC1_VIDEO_MODULE_stream_init(void *mem, PMT_SECTION *pmt, int idx)
{
   return (pmt->elem[idx]->stream_type == STREAM_TYPE_WMV_DATA);
}

/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PES_PKT that needs handling.
                        Set pes->stream_id to zero to let the PES
                        module know to expect the start of a new
                        PES_PKT. */
void VC1_VIDEO_MODULE_process_pkt(void *mem, void *pes)
{
   VC1_video_module_data *data = (VC1_video_module_data *)mem;
   PES_PKT               *pkt  = (PES_PKT *)pes;

#ifndef _SPEED_TEST_
   if (PES_PTS_FLAG(pkt))
      data->pes_pts = (pkt->PTS) ? pkt->PTS : 1;
#endif

   data->wmv.payload_len = (u32)pkt->payload_len16;
   data->wmv.payload     = pkt->payload;

   while (data->wmv.payload_len) {
      switch (data->wmv.state & WMV_STATE_MASK) {
      case WMV_next_guid_STATE:
         WMV_next_guid(&(data->wmv));
         break;
      case WMV_GUID_FILE:
         WMV_read_file(&(data->wmv));
         break;
      case WMV_GUID_STREAM:
         WMV_read_stream(&(data->wmv), &(data->seq));
         if ( ((data->wmv.state & WMV_STATE_MASK) != WMV_GUID_STREAM)&&
              (data->wmv.stream_id) ) {
            init_display(data);
         }
         break;
      case WMV_GUID_DATA:
         WMV_read_data(&(data->wmv));
         break;
      case WMV_FRAME_FRAG_STATE:
         if (data->wmv.payload_len < data->wmv.WD_frag_size)
            data->seq.payload_len = data->wmv.payload_len;
         else {
            data->seq.payload_len = data->wmv.WD_frag_size;
            data->wmv.state       = WMV_END_FRAME_STATE;
         }
         data->seq.payload       = data->wmv.payload;
         data->wmv.payload      += data->seq.payload_len;
         data->wmv.payload_len  -= data->seq.payload_len;
         data->wmv.WD_frag_size -= data->seq.payload_len;

         if (!data->wmv.WD_frag_offset) {
            /* start frame */
            data->wmv.WD_frag_offset++;
            data->seq.state = VC1_VIDEO_read_PIC_LAYER1_STATE;
         }
         if ( (!data->wmv.WD_frag_size)&&
              (data->wmv.frame_size == data->wmv.frag_offset) ) {
            /* end frame */
            data->seq.payload_len++;
         }
         simple_main_decode(data);
         break;
      case WMV_skip_guid_STATE:
      default:
         WMV_skip_guid(&(data->wmv));
      }
   }

   pkt->payload       = data->wmv.payload;
   pkt->payload_len16 = (u16)data->wmv.payload_len;
}


static void simple_main_decode(VC1_video_module_data *data)
{
   SEQ_LAYER *seq = &(data->seq);
   PIC_LAYER *pic = &(data->pic);

   while (seq->payload_len) {
      switch (seq->state & VC1_VIDEO_STATE_MASK) {
      case VC1_VIDEO_read_PIC_LAYER1_STATE:
         VC1_VIDEO_read_PIC_LAYER1(seq, pic);
         if (seq->state == VC1_VIDEO_read_PIC_LAYER2_STATE)
            new_frame(data);
         break;
      case VC1_VIDEO_read_PIC_LAYER2_STATE:
         VC1_VIDEO_read_PIC_LAYER2(seq, pic);
         break;
      case VC1_VIDEO_read_I_MACROBLOCK_STATE:
         VC1_VIDEO_read_I_MACROBLOCK(seq, pic);
         break;
      case VC1_VIDEO_read_P_MACROBLOCK_STATE:
         VC1_VIDEO_read_P_MACROBLOCK(seq, pic);
         break;
      default:
         seq->payload_len = 0;
      }
   }
}

static void init_display(VC1_video_module_data *data)
{
   data->pic.last_frame = 0;
   data->pic.ref_blkF   = NULL;
   data->pic.ref_blkB   = NULL;
#ifndef OUTPUT_BLOCK_TRACE
   if ( (DISPLAY_init((u32)data->seq.wid,      (u32)data->seq.hgt,
                      (u32)data->seq.disp_wid, (u32)data->seq.disp_hgt))&&
        (data->pic.frame_wid = DISPLAY_frame_wid()) ) {
      u32 bytes_needed = FRAME_420_SIZE(data->pic.frame_wid, (u32)data->seq.hgt);

      bytes_needed       <<= 1;  /* we need 2 frames */
      data->pic.ref_frameF = DISPLAY_ref_buf(bytes_needed);
      if (!data->pic.ref_frameF) {
# ifdef _SHOW_ERRORS_
         printf("VC1_VIDEO_MODULE: "
                "Could not allocate memory for reference frames.\n");
# endif
         return;
      }
      data->pic.luma_size  = bytes_needed /  3;
      data->pic.frame_size = bytes_needed >> 1;
      data->pic.ref_frameB = data->pic.ref_frameF + data->pic.frame_size;
   }
   else {
      VC1_VIDEO_MODULE_done(data);
# ifdef _SHOW_ERRORS_
      printf("VC1_VIDEO_MODULE: "
             "Could not initialize display.\n");
# endif
   }
#else
   data->pic.frame_wid  = (u32)data->seq.wid;
   data->pic.frame_size = FRAME_420_SIZE((u32)data->seq.wid,
                                         (u32)data->seq.hgt);
   data->pic.luma_size  = (u32)data->seq.wid * (u32)data->seq.hgt;
   data->pic.ref_frameF = (u8 *)10000000;
   data->pic.ref_frameB = (u8 *)20000000;
#endif
   data->output_delay  = MPEG2_SYS_CLK_FREQ * (CLK_VAL)data->seq.framerate_den;
   data->output_delay /= (CLK_VAL)data->seq.framerate_num;
}

static void new_frame(VC1_video_module_data *data)
{
   SEQ_LAYER *seq = &(data->seq);
   PIC_LAYER *pic = &(data->pic);

   if (!pic->ref_blkF) {
      if (VC1_PIC_TYPEp(pic) != VC1_I_FRAME) {
         seq->state = VC1_VIDEO_frame_completed_STATE;
         return;  /* start of sequence must be an I frame */
      }

      pic->pred_frame = pic->ref_frameB;
      pic->last_frame = VC1_I_FRAME;
      data->cur_pts   = MPEG2_prog_clk(data->offset) + data->output_delay;
   }
   else if (!pic->last_frame)
      goto NEXT_FRAME;
   else {
      /* this is the start of a new frame */
#ifndef OUTPUT_BLOCK_TRACE
      if (!(pic->last_frame & VC1_B_FRAME)) {
         MV_COPY_FRAME(DISPLAY_get_next_frame(),
                       pic->ref_frameB,
                       pic->frame_size);
      }
      else
         goto NEXT_FRAME;  /* unable to decode B frames */

      /* output the next frame */
      if ( (DISPLAY_next_frame_done(data->cur_pts, data->offset))&&
           (data->sys_data->a_sync)&&(data->sys_data->v_skip <= 0) )
         data->sys_data->v_skip = 1;   /* drop frames, since we are behind */
      AV_SYNC(data->cur_pts, data->pes_pts, data->sys_data->a_sync)

NEXT_FRAME:
#endif
      /* store new frame type, and get next frame */
      pic->last_frame = VC1_PIC_TYPEp(pic);
      if (pic->last_frame != VC1_B_FRAME) {
         /* ref_frameB becomes the new ref_frameF */
         pic->pred_frame = pic->ref_frameF;
         pic->ref_frameF = pic->ref_frameB;
         pic->ref_frameB = pic->pred_frame;
      }
      else
         seq->state = VC1_VIDEO_frame_completed_STATE;

#ifndef _SPEED_TEST_
      data->pes_pts          = 0;
      data->sys_data->v_pos += data->output_delay;
      if (data->sys_data->v_skip > 0) {
         data->sys_data->v_skip -= data->output_delay;
         if (data->sys_data->v_skip <= 0) {
            if (pic->last_frame != VC1_I_FRAME)
               data->sys_data->v_skip = 1;   /* wait for an I frame */
            else
               data->cur_pts = MPEG2_prog_clk(data->offset);
         }
         if (data->sys_data->v_skip > 0) {
            pic->last_frame = VC1_SKIP_FRAME;
            seq->state      = VC1_VIDEO_frame_completed_STATE;
         }
      }

      if (data->sys_data->command == CMD_SLOW_FWD)
         data->cur_pts += data->output_delay << CMD_SLOW_SHIFT;
      else
         data->cur_pts += data->output_delay;
#endif
   }

   /* reset blk pointers */
   pic->ref_blkF = pic->ref_frameF;
   pic->ref_blkB = pic->ref_frameB;
   pic->pred_blk = pic->pred_frame;
}
